from django_nose.tools import *
from assertions import BeExactlyAssertion as be_exactly, \
  BeValidAssertion as be_valid, ContainAssertion as contain, \
  EqualAssertion as equal, EqualAssertion as be, \
  HaveSelectorAssertion as have_selector, HaveTextAssertion as have_text, \
  RespondToAssertion as respond_to

from expectations import ChangeExpectation as change

from base import RedRoverHelper, RedRoverLiveTest, RedRoverTest, before, \
  describe
from expect import Expect as expect
from runner import RedRoverRunner
from subject import BROWSER as page


__all__ = [
  'RedRoverHelper', 'RedRoverLiveTest', 'RedRoverRunner', 'RedRoverTest',
  'before', 'describe', 'expect', 'page',

  # RedRover assertions
  'be', 'be_exactly', 'be_valid', 'contain', 'equal', 'have_selector',
  'have_text', 'respond_to',

  # RedRover expectations
  'change',

  # Nose assertions
  'assert_true', 'assert_equal', 'assert_template_not_used',
  'assert_set_equal', 'assert_almost_equals', 'assert_greater_equal',
  'assert_sequence_equal', 'assert_not_equals', 'assert_is_none',
  'assert_raises', 'assert_almost_equal', 'assert_num_queries',
  'assert_not_in', 'assert_regexp_matches', 'assert_ok',
  'assert_not_almost_equals', 'assert_mail_count', 'assert_h_t_m_l_equal',
  'assert_is_instance', 'assert_not_regexp_matches', 'assert_contains',
  'assert_raises_regexp', 'assert_raises_message', 'assert_template_used',
  'assert_tuple_equal', 'assert_false', 'assert_field_output',
  'assert_not_contains', 'assert_less_equal', 'assert_list_equal',
  'assert_multi_line_equal', 'assert_items_equal', 'assert_form_error',
  'assert_equals', 'assert_not_is_instance', 'assert_is_not',
  'assert_h_t_m_l_not_equal', 'assert_dict_contains_subset', 'assert_greater',
  'assert_is', 'assert_redirects', 'assert_not_equal', 'assert_queryset_equal',
  'assert_dict_equal', 'assert_code', 'assert_in', 'assert_is_not_none',
  'assert_not_almost_equal', 'assert_less']



class ExpectationError(AssertionError):
  pass


def _expectation(_expected_outcome, _action_class, *_args, **_kwargs):

  class Expectation(object):

    def __init__(self, action_class, *args, **kwargs):
      self.action = action_class(*args, **kwargs)

    def __enter__(self):
      return self.action.enter()

    def __exit__(self, type, value, traceback):
      if self.action.exit() != _expected_outcome:
        raise ExpectationError(self.action.message)
        return False
      return True

  return Expectation(_action_class, *_args, **_kwargs)


class Expect(object):

  @classmethod
  def to(cls, action_class, *args, **kwargs):
    return _expectation(True, action_class, *args, **kwargs)

  @classmethod
  def not_to(cls, action_class, *args, **kwargs):
    return _expectation(False, action_class, *args, **kwargs)



from django.core.urlresolvers import NoReverseMatch, reverse
from django.db.models import Model as DjangoModel


def get_url(obj, *args, **kwargs):
  """Try to infer a URL from an ambiguous object."""
  if isinstance(obj, DjangoModel):
    return obj.get_absolute_url()
  obj = str(obj)
  if obj.startswith('http'):
    return obj
  try:
    return reverse(obj, *args, **kwargs)
  except NoReverseMatch:
    return obj
  except Exception, msg:
    raise RuntimeError(msg)



from base import BaseExpectation


class ChangeExpectation(BaseExpectation):
  """Determine if the result of a callable changes."""

  def __init__(self, subject, by=None):
    if callable(subject):
      self.subject = subject
    else:
      raise RuntimeError(
        "The subject of a change expectation must be a callable.")
    self.delta = by

  def enter(self):
    self.original_value = self.subject()
    return self.original_value

  def exit(self):
    self.new_value = self.subject()
    if self.delta:
      self.passes = abs(self.original_value - self.new_value) == self.delta
    else:
      self.passes = self.original_value != self.new_value
    return self.passes

  @property
  def message(self):
    if self.passes:
      if self.delta:
        msg = '{subject} changed by {delta}'
      else:
        msg = '{subject} changed'
    else:
      if self.delta:
        msg = '{subject} did not change by {delta} (it changed by {actual})'
      else:
        msg = '{subject} did not change'

    return msg.format(
      subject=repr(self.subject),
      delta=repr(self.delta),
      actual=abs(self.original_value - self.new_value))



import functools
from types import FunctionType

from django.test import LiveServerTestCase, TestCase
import splinter

from subject import get_splinter_actions, _subject, BROWSER


def _get_extra_context(instance, extra_context={}):
  subject_name = getattr(instance, 'subject', None)

  if subject_name:
    extra_context = {
      'it': _subject(instance)(subject_name),
      'its': _subject(instance, parent_name=subject_name)}

    if subject_name == BROWSER:
      extra_context.update(get_splinter_actions(instance))

  return extra_context


def before(func):
  """
  The only way to get `setUp` to behave the way we want is to decorate
  it.  So might as well make it look intentional :)

  """
  @functools.wraps(func)
  def setUp(instance, *args, **kwargs):
    func.__globals__.update(_get_extra_context(instance))
    return func(instance, *args, **kwargs)

  return setUp


def describe(func):
  """
  This more or less is the "magic" of RedRover.  With the `describe`
  decorator, a test is equipped with the keywords 'it', 'its', and
  if applicable, 'visit'.

  """
  @functools.wraps(func)
  def test_func(instance, *args, **kwargs):
    func.__globals__.update(_get_extra_context(instance))
    return func(instance, *args, **kwargs)

  return test_func


def _is_protected(item):
    """Filter out things we shouldn't be messing with."""
    return ('__' in item or item in dir(LiveServerTestCase))


def _redrover_klass():
  """
  Prefix test methods with 'test_' so they are discoverable by the
  test runner.

  """
  class DiscoverableTests(type):

    def __new__(cls, name, bases, dct):
      for attr, value in dct.iteritems():
        if not _is_protected(attr) and isinstance(value, FunctionType):
          value.__name__ = 'test_%s' % value.__name__
      return super(DiscoverableTests, cls).__new__(cls, name, bases, dct)

  return DiscoverableTests


class RedRoverTest(TestCase):

  __metaclass__ = _redrover_klass()


class RedRoverLiveTest(LiveServerTestCase):

  def __init__(self, *args, **kwargs):
    if getattr(self, 'subject', None) == BROWSER:
      setattr(self, BROWSER, splinter.Browser('zope.testbrowser'))
      self.setUp.__globals__.update(get_splinter_actions(self))
    super(RedRoverLiveTest, self).__init__(*args, **kwargs)

  __metaclass__ = _redrover_klass()


class RedRoverHelper(object):

  def __init__(self, parent):
    for i in dir(parent):
      if not i.startswith('__'):
        setattr(self, i, getattr(parent, i))
    if self.subject == BROWSER:
      self.entry.__globals__.update(get_splinter_actions(parent))
      self.exit.__globals__.update(get_splinter_actions(parent))

  def __enter__(self):
    return self.entry()

  def __exit__(self, exc_type, exc_value, traceback):
    if exc_type:
      return False
    return self.exit()

  def entry(self):
    pass

  def exit(self):
    pass



from urlparse import urlparse

from base import BaseAssertion


class HaveSelectorAssertion(BaseAssertion):
  """Determine if given selector exists in a splinter Browser document."""

  def __init__(self, browser, selector, **kwargs):
    if not hasattr(browser, 'html'):
      raise ValueError(
        '`have_selector` must be called on a page/Browser instance.')
    self.url = urlparse(browser.url).path
    self.selector = selector
    self.extras = kwargs
    elements = browser.find_by_css(selector)
    # Special requests
    for k in kwargs:
      elements = [el for el in elements if getattr(el, k).strip() == kwargs[k]]

    self.passes = bool(elements)

  @property
  def message(self):
    if self.passes:
      msg = '{selector} selector{extras} exists in {url}'
    else:
      msg = '{selector} selector{extras} does not exist in {url}'

    if self.extras:
      extras = ' with the %s' % ', and the '.join(
        ['%s "%s"' % (k, self.extras[k]) for k in self.extras])
    else:
      extras = ''

    return msg.format(
      selector=repr(self.selector),
      url=self.url,
      extras=extras)



from httplib import InvalidURL
from urlparse import urlparse

from splinter.exceptions import ElementDoesNotExist
from zope.testbrowser.browser import Browser as ZopeTestBrowser
from utils import get_url


class ElementNotFound(AssertionError):
  pass


BROWSER = '_browser'


def _splinter_action(_action, _parent):
  """
  All the various splinter actions we support.  (and a few others)

  """
  _browser = getattr(_parent, BROWSER)

  # Finders

  def find_button(query):
    # Grab all buttons
    buttons = _browser.find_by_tag('button')
    # First try to find the right one matching text
    matching_text = filter(lambda e: e.text.strip() == query, buttons)
    if matching_text:
      return matching_text[0]
    # Try by CSS selector
    buttons = _browser.find_by_css(query)
    if buttons:
      return buttons.first
    raise ElementNotFound('Could not find a button matching "%s"' % query)

  def find_link(query):
    # First we'll try by text, then by CSS selector
    links = _browser.find_link_by_text(query)
    if links:
      return links.first
    links = _browser.find_by_css('a').find_by_css(query)
    if links:
      return links.first
    raise ElementNotFound('Could not find a link matching "%s"' % query)

  def find_input(query):
    #
    # NO LONGER USED
    #
    # Try by label text first
    labels = _browser.find_by_tag('label')
    matching_text = filter(lambda e: e.text == query, labels)
    if matching_text:
      input_id = matching_text[0]._element.for_element.attrib['id']
      try:
        return _browser.find_by_id(input_id).first
      except ElementDoesNotExist:
        raise ElementNotFound(
          'Found label "%s", but no matching input' % query)
    # Try by name
    try:
      inputs = _browser.find_by_name(query)
    except LookupError:
      # Zope Testbrowser throws this if the query is not a valid name.
      inputs = None
    if inputs:
      return inputs.first
    # Try by CSS selector
    inputs = _browser.find_by_css(query)
    if inputs:
      return inputs.first
    raise ElementNotFound('Could not find an input matching "%s"' % query)

  # Actions

  def fill_in(query, value):
    try:
      control = _browser._browser.getControl(query)
    except LookupError:
      try:
        control = _browser._browser.getControl(name=query)
      except LookupError:
        control = None

    if control:
      control.value = value
    else:
      input_field = find_input(query)
      input_field.fill(value)

  def click_on(query):
    try:
      # First try it the easy way (if the button is actually a
      # submit input).
      control = _browser._browser.getControl(query)
    except LookupError:
      button = find_button(query)
      try:
        control = _browser._browser.getControl(
          name=button._element.attrib['name'])
      except KeyError:
        raise KeyError(
          'Buttons must have a unique `name` attribute to be used.')
    return control.click()

  def select(query, option):
    try:
      control = _browser._browser.getControl(query)
    except LookupError:
      control = _browser._browser.getControl(name=query)
    control.value = [option]

  def visit(location, *args, **kwargs):
    try:
      url = get_url(location, *args, **kwargs)
      if url.startswith('/'):
        base_url = _parent.live_server_url
        return _browser.visit('%s%s' % (base_url, url))
      return _browser.visit(location)
    except:
      raise InvalidURL(
        'Could not determine a reliable URL from "%s"' % location)

  def click_link(*args, **kwargs):
    control = _browser._browser.getLink(*args, **kwargs)
    return control.click()

  return locals().get(_action)


def get_splinter_actions(instance):
  """
  A dictionary of all supported Splinter actions plus a few of my own
  set up against the given test-class instance.

  """
  finders = ['find_button', 'find_link', 'find_input']
  actions = ['click_link', 'click_on', 'fill_in', 'select', 'visit']

  all_actions = {k: _splinter_action(k, instance) for k in finders + actions}
  all_actions['current_path'] = _subject(
    instance._browser, modifier=lambda url: urlparse(url).path)('url')
  return all_actions


def _subject(instance, parent_name=None, modifier=None):
  """
  The Subject object references a particular attribute and allows the
  developer to check it for various behaivors and conditions.

  """
  class Subject(object):

    def __init__(self, name):
      self.name = name
      self.instance = instance  # For debugging

    def should(self, cls, *args, **kwargs):
      assertion = cls(self.value, *args, **kwargs)
      if not assertion.passes:
        raise AssertionError(assertion.message)
      return True

    def should_not(self, cls, *args, **kwargs):
      assertion = cls(self.value, *args, **kwargs)
      if assertion.passes:
        raise AssertionError(assertion.message)
      return True

    @property
    def value(self):
      if parent_name:
        parent = getattr(instance, parent_name)
        if self.name == 'path':
          if isinstance(getattr(parent, BROWSER, None), ZopeTestBrowser):
            _value = urlparse(parent.url).path
        else:
          _value = getattr(parent, self.name)
      else:
        _value = getattr(instance, self.name)
      if modifier:
        _value = modifier(_value)
      return _value

    def __str__(self):
      return str(self.value)

    def __repr__(self):
      return self.__str__()

    def __lt__(self, other):
      return self.value < other

    def __le__(self, other):
      return self.value <= other

    def __eq__(self, other):
      return self.value == other

    def __ne__(self, other):
      return self.value != other

    def __gt__(self, other):
      return self.value > other

    def __ge__(self, other):
      return self.value >= other

  return Subject



from base import BaseAssertion


class ContainAssertion(BaseAssertion):
  """Determine if one object exists within another."""

  def __init__(self, subject, other):
    self.subject = subject
    self.other = other
    self.passes = other in subject

  @property
  def message(self):
    if self.passes:
      msg = '{other} is in {subject}'
    else:
      msg = '{other} is not in {subject}'

    return msg.format(
      subject=repr(self.subject),
      other=repr(self.other))
